//
// Title:   CoTeXTool.cpp
// Created: Mon Mar 17 19:00:00 2003
// Authors: Clment Bichel, Blanca B. Borro Escribano, Tamy Boubekeur, Elisabeth Brunet
//
// Copyright (c) 2003, Funky CoTeX B Group, University of Bordeaux
//
// This file is part of the CoTeX plugin for PointShop 3D
// See http://www.cotex.fr.st for more information.
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later
// version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General
// Public License along with this library; if not, write to the
// Free Software Foundation, Inc., 59 Temple Place, Suite 330,
// Boston, MA 02111-1307 USA
// 
// Contact info@pointshop3d.com if any conditions of this
// licensing are not clear to you.
//

#include <qstring.h>
#include <qtoolbutton.h>
#include <qtoolbar.h>
#include <qcursor.h>
#include <qevent.h>
#include <qpixmap.h>
#include <qwhatsthis.h>
#include <qwidget.h>
#include <qpainter.h>
#include <qcolor.h>
#include <qlineedit.h>
#include <qlistbox.h>
#include <qbuttongroup.h>
#include <qtextstream.h>
#include <qfile.h>
#include <qpushbutton.h>
#include <assert.h>


#include <iostream>
#include <Scene.h>
#include <RendererManager.h>
#include <ToolInterface.h>
#include <StatusBar.h>




#include "CoTeXTool.h"

#include "../ui/CoTeXConfigurationWidget.h"
#include "ColorCoTeXTree/ColorCoTeXTree.h"
#include "Primitives/SphereManager.h"
#include "Primitives/ConvPointManager.h"
#include "Primitives/RbfManager.h"
#include "Primitives/BlockManager.h"
#include "Operators/OperatorUnionManager.h"
#include "Textures/SimpleColorTexture.h"
#include "Operators/OperatorUnion.h"
#include "CoTeXIO/CoTeXIO.h"

// ******************
// 'What's this' text
// ******************

static const char *whatsThisText = "<img source=\"CoTeX\"> "
"Click this button to activate the CoTeX tool"
"<ul><li>A Constructive Texturing Tool</li>"
"<li>Partition Edition</li></ul>";

// ****************
// CoTeX Tool class
// ****************

using namespace std;

CoTeXTool::CoTeXTool () {
  toolCursor = new QCursor (PointingHandCursor);
  cotexConfigurationWidget = new CoTeXConfigurationWidget (&textures);
  primitiveManagerMap.insert ("sphere", SphereManager::getInstance ());
  primitiveManagerMap.insert ("convpoint", ConvPointManager::getInstance ());
  primitiveManagerMap.insert ("rbf", RbfManager::getInstance ());
  primitiveManagerMap.insert ("block", BlockManager::getInstance ());
  operatorManagerMap.insert ("union", OperatorUnionManager::getInstance ());
  lastPrimitiveCreated = NULL;
  objectRemoved = NULL;
  selected = false;
  
  // Connection with userInterface
  connect (cotexConfigurationWidget, SIGNAL (exportHFSignal (QString)),
		   this, SLOT (handleExportHF (QString)));
  
  connect (cotexConfigurationWidget, SIGNAL (loadSignal (QString)),
		   this, SLOT (handleLoad (QString)));

  connect (cotexConfigurationWidget, SIGNAL (saveSignal (QString)),
		   this, SLOT (handleSave (QString)));

  connect (cotexConfigurationWidget, SIGNAL (resetSignal ()),
		   this, SLOT (handleReset ()));

  // Connection for scene's signals
  Scene * scene = Scene::getInstance ();
  connect (scene, SIGNAL (activeObjectChanged ()), this, 
		   SLOT (handleActiveObjectChanged ()));
  connect (scene, SIGNAL (activeObjectAboutToChange ()), this, 
		   SLOT (handleActiveObjectAboutToChange ()));
  connect (scene, SIGNAL (objectRemoved ()), this, 
		   SLOT (handleObjectRemoved ()));
}

CoTeXTool::~CoTeXTool () {
  delete toolCursor;
  delete cotexConfigurationWidget;
  cleanUp ();
}

void CoTeXTool::cleanUp () {
  // Destruction of textures
  TextureInterface::TextureMap::Iterator itTex;
  for (itTex = textures.begin (); itTex != textures.end (); itTex++)
    delete *itTex;

  // Destruction of primitives managers
  SphereManager::destroyInstance ();
  ConvPointManager::destroyInstance ();
  RbfManager::destroyInstance ();
  primitiveManagerMap.clear ();

  // Destruction of operators managers
  OperatorUnionManager::destroyInstance ();
  operatorManagerMap.clear ();

  // Destruction of of CoTeXTrees and CoTeX's primtives Objects.
  Scene * scene = Scene::getInstance ();
  Object * itO;
  for (itO = scene->getFirstObject (); itO != NULL; itO = scene->getNextObject ()) {
    delete coTeXTreeMap[itO];
    destroyCoTeXObjects (itO);
  }
}


QToolButton *CoTeXTool::addButtonToToolBar (QToolBar *toolBar) {
  QIconSet iconSet;
  // relative path from the executable
  QPixmap  normalIcon ("Resources/CoTeXToolIcon.png"),
	disabledIcon ("Resources/CoTeXToolIcon_disabled.png");
  
  iconSet.setPixmap (normalIcon, QIconSet::Automatic, QIconSet::Normal);
  iconSet.setPixmap (disabledIcon, QIconSet::Automatic, QIconSet::Disabled);
  toolButton = new QToolButton (iconSet, "CoTeX Tool", "Paint with CoTeX primitive",
								0, 0, toolBar, "CoTeX Tool");
  toolButton->setToggleButton (true);
  toolButton->setEnabled (this->checkToolAvailability());
  
  // add a short description to this button
  QWhatsThis::add (toolButton, whatsThisText);
  
  // make the icon visible in the 'What's this'-text
  QMimeSourceFactory::defaultFactory()->setPixmap ("cotex", normalIcon);

  return toolButton;

}

QToolButton *CoTeXTool::getToolButton () {
  return toolButton;
}

QCursor *CoTeXTool::getToolCursor () const {
  return toolCursor;
}

QWidget *CoTeXTool::getToolConfigurationWidget () const {
  return cotexConfigurationWidget;
}

bool CoTeXTool::isAvailable () const {
  return this->checkToolAvailability();
}

void CoTeXTool::selectTool () {
  // Create CoTeXTree if Active Object don't has any
  Scene * scene = Scene::getInstance ();
  Object * activeObject = scene->getActiveObject ();
  
  // the implicit call to checkToolAvailability ensure that the
  // active object isn't NULL.

  selected = true;
  if (coTeXTreeMap[activeObject] == NULL) 
    coTeXTreeMap.insert (activeObject, new ColorCoTeXTree ());
  
  // Put object's primitives for vizualisation
  showCoTeXObjects ();
  cerr << "CoTeX : Selected" << endl;
}

void CoTeXTool::deselectTool () {
  // Remove object's primitives of scene for optimized vizualisation
  hideCoTeXObjects ();  
  selected = false;
  cerr << "CoTeX : Deselected" << endl;
}


////////////// HANDLERS ///////////////////////////////////////////////////////

void CoTeXTool::handleMousePressEvent (QMouseEvent *me) {
  QString typeOfPrimitive = 
	cotexConfigurationWidget->getCurrentPrimitiveTypeName ();
  PrimitiveManagerInterface * currentPrimitiveManager = 
	primitiveManagerMap[typeOfPrimitive];
   try {
	currentPrimitiveManager->handleMousePressEvent (me, cotexConfigurationWidget);
  } catch (PrimitiveManagerException e) {
	StatusBar * statusBar = StatusBar::getInstance ();
	statusBar->showMessage (e.getMessage (), 0);
  }
}


void CoTeXTool::handleMouseMoveEvent (QMouseEvent *me) {
  QString typeOfPrimitive = 
	cotexConfigurationWidget->getCurrentPrimitiveTypeName ();
  PrimitiveManagerInterface * currentPrimitiveManager = 
	primitiveManagerMap[typeOfPrimitive];
  try {
	currentPrimitiveManager->handleMouseMoveEvent (me);
  } catch (PrimitiveManagerException e) {
	StatusBar * statusBar = StatusBar::getInstance ();
	statusBar->showMessage (e.getMessage (), 0);
  }
}


void CoTeXTool::handleMouseReleaseEvent (QMouseEvent *me) {
  QString typeOfPrimitive = 
	cotexConfigurationWidget->getCurrentPrimitiveTypeName ();
  PrimitiveManagerInterface * currentPrimitiveManager = 
	primitiveManagerMap[typeOfPrimitive];
  
  try {
	currentPrimitiveManager->handleMouseReleaseEvent (me);
  } catch (PrimitiveManagerException e) {
	StatusBar * statusBar = StatusBar::getInstance ();
	statusBar->showMessage (e.getMessage (), 0);
  }
  
  PrimitiveInterface *  newPrimitive = currentPrimitiveManager->getCurrentPrimitive (); 
  if (newPrimitive != NULL) {
    
    Scene * scene = Scene::getInstance ();
    Object * activeObject = scene->getActiveObject ();
    
    // *********************
    // adding object to list
    // *********************
    Object * newObject = currentPrimitiveManager->getCurrentObject (); 
    assert (newObject != NULL);
    if (objectObjects[activeObject] == NULL)
      objectObjects.insert (activeObject, 
			    new PrimitiveManagerInterface::PrimitiveObjectMap);
    objectObjects[activeObject]->insert (newPrimitive, newObject);
  
	// ***************************
    // Generating the new operator
    // ***************************
	QString typeOfOperator = 
	  cotexConfigurationWidget->getCurrentOperatorTypeName ();
	OperatorManagerInterface * currentOperatorManager = 
	  operatorManagerMap[typeOfOperator];
	OperatorInterface * newOperator = currentOperatorManager->createOperator ();
	  
    // *********************************************
    // update and apply the tree to the activeObject
    // *********************************************
	
    coTeXTreeMap[activeObject]->addPrimitive (newPrimitive, newOperator); 
    coTeXTreeMap[activeObject]->apply (activeObject);
    lastPrimitiveCreated = newPrimitive;
    lastObjectTextured = activeObject;
  }
}

void CoTeXTool::handleKeyPressEvent (QKeyEvent *ke) { 
  switch (ke->key ()) {
  case Key_Delete :
	if (lastPrimitiveCreated != NULL) {
	  Scene * scene = Scene::getInstance ();
	  bool autoDeleteState = scene->isAutoDelete ();
	  
	  delete coTeXTreeMap[lastObjectTextured]->removePrimitive (lastPrimitiveCreated);
	  scene->setAutoDelete (true);
	  
	  scene->removeObject ((*objectObjects[lastObjectTextured])[lastPrimitiveCreated], false);
	  objectObjects[lastObjectTextured]->remove (lastPrimitiveCreated);
	  scene->setAutoDelete (autoDeleteState);
	  coTeXTreeMap[lastObjectTextured]->apply (lastObjectTextured);
	  lastPrimitiveCreated = NULL;
	}
	break;
  case Key_F1 : 
	exportHyperFun ("quicksave.hf");
  	break;
  default:
	break;
  }
}

void CoTeXTool::handleKeyReleaseEvent (QKeyEvent *ke) {
  ke->ignore ();
}

///////////// PUBLIC SLOTS ///////////////////////////////////////////////

void CoTeXTool::handleRendererWidgetRepainted () {
  // the RenderWidget has repainted its content, so here this tool
  // can draw its content with a QPainter, if necessary
}

///////////// PRIVATE SLOTS ///////////////////////////////////////////////

// *******************
// Conections with UI
// *******************

void CoTeXTool::handleExportHF (QString filename) {
  if (!filename.isNull ())              
	exportHyperFun (filename);
}

void CoTeXTool::handleSave (QString filename) {
  if (!filename.isNull ()) {
	Scene * scene = Scene::getInstance ();
	Object * activeObject = scene->getActiveObject ();
	CoTeXIO::save (&textures, coTeXTreeMap[activeObject], filename);
  }
}

void CoTeXTool::handleLoad (QString filename) {
  if (!filename.isNull ()) {
	Scene * scene = Scene::getInstance ();
	Object * activeObject = scene->getActiveObject ();
	
	handleReset ();
	
	// creation of the object map for primitives 
	if (objectObjects[activeObject] == NULL)
      objectObjects.insert (activeObject, 
							new PrimitiveManagerInterface::PrimitiveObjectMap);
	
	try {
	CoTeXIO::load (coTeXTreeMap[activeObject], activeObject, cotexConfigurationWidget,
				   objectObjects[activeObject], &primitiveManagerMap, 
				   &operatorManagerMap, filename);
	coTeXTreeMap[activeObject]->apply (activeObject);
	} catch (CoTeXIOException e) {
	  StatusBar * statusBar = StatusBar::getInstance ();
	  statusBar->showMessage (e.getMessage (), 0);
	}
  }
}

void CoTeXTool::handleReset () {
  Scene * scene = Scene::getInstance ();
  Object * activeObject = scene->getActiveObject ();
  
  if (activeObject != NULL) {
	// destroy the old tree
	delete coTeXTreeMap[activeObject];
	coTeXTreeMap.remove (activeObject);
	destroyCoTeXObjects (activeObject);
	
	// place a new tree for this object
	ColorCoTeXTree * newTree =  new ColorCoTeXTree ();
	coTeXTreeMap.insert (activeObject, newTree);
	newTree->apply (activeObject);
  }
}

// *************************
// Conections with PointShop
// *************************

void CoTeXTool::handleActiveObjectAboutToChange () {
  // only used if there's an objectRemoved signal.
  objectRemoved = Scene::getInstance ()->getActiveObject ();
  if (objectRemoved != NULL)
	hideCoTeXObjects ();
}

void CoTeXTool::handleActiveObjectChanged () {
  Scene * scene = Scene::getInstance ();
  Object * activeObject = scene->getActiveObject ();
  if ((activeObject) != NULL && (coTeXTreeMap[activeObject] == NULL)) {
    coTeXTreeMap.insert (activeObject, new ColorCoTeXTree ());
    cerr << "CoTeX : new ColorCoTeXTree for the active Object." << endl;
	showCoTeXObjects ();
  }  
}

void CoTeXTool::handleObjectRemoved () {
  // When a scene is loaded, an objectRemoved Signal is emitted
  // but there was no object so we must test.
  if (objectRemoved != NULL) {
    delete coTeXTreeMap[objectRemoved];
    coTeXTreeMap.remove (objectRemoved);
    destroyCoTeXObjects (objectRemoved);
  }
  objectRemoved = NULL;
}
  
////////////// PRIVATE METHODS ///////////////////////////////////////////

bool CoTeXTool::checkToolAvailability () const {
  return RendererManager::getInstance ()->getMainRenderer ()->isInteractive () &&
	(Scene::getInstance ()->getNofObjects () > 0);
}

void CoTeXTool::showCoTeXObjects (Object * object) {
  Scene * scene = Scene::getInstance ();
  PrimitiveManagerInterface::PrimitiveObjectMap::Iterator it;  
  
  if (objectObjects[object] != NULL)
    for (it = objectObjects[object]->begin (); 
	 it != objectObjects[object]->end (); 
	 it++)
      scene->addObject (*it, false, false);
  scene->emitRenderFrameSignal ();
}

void CoTeXTool::showCoTeXObjects () {
  Scene * scene = Scene::getInstance ();
  Object * activeObject = scene->getActiveObject ();
  
  showCoTeXObjects (activeObject);
}

void CoTeXTool::hideCoTeXObjects (Object * object) {
  removeCoTeXObjects (object, false);
}

void CoTeXTool::hideCoTeXObjects () {
  Scene * scene = Scene::getInstance ();
  Object * activeObject = scene->getActiveObject ();
  
  hideCoTeXObjects (activeObject);
}

void CoTeXTool::destroyCoTeXObjects (Object * object) {
  removeCoTeXObjects (object, true);
  cerr << "CoTeX : destruction of CoTeX's Primitives" << endl;
  delete objectObjects[object];
  objectObjects.remove (object);
}  

void CoTeXTool::removeCoTeXObjects (Object * object, bool autoDelete) {
   Scene * scene = Scene::getInstance ();
   bool autoDeleteState = scene->isAutoDelete ();
  
   scene->setAutoDelete (autoDelete);
   PrimitiveManagerInterface::PrimitiveObjectMap::Iterator it; 
   if (objectObjects[object] != NULL)
     for (it = objectObjects[object]->begin (); 
	  it != objectObjects[object]->end (); 
	  it++) 
       if (selected)
	 scene->removeObject (*it, false);
       else if (!selected && autoDelete)
	 delete *it;
          
   scene->emitRenderFrameSignal ();
   scene->setAutoDelete (autoDeleteState);
}

bool CoTeXTool::exportHyperFun (QString fileName) {
  Scene * scene = Scene::getInstance ();
  Object * activeObject = scene->getActiveObject ();
  StatusBar * statusBar = StatusBar::getInstance (); // informations for users
  QFile file (fileName);
  
  statusBar->showMessage ("Exporting Current Tree to " + fileName + ", wait ..."  , 0);
  file.open (IO_WriteOnly);         
  QTextStream stream (&file);
  stream.flags (QTextStream::fixed); // HyperFun doesn't accept scientif expression for reals
  CoTeXTreeInterface * tree = coTeXTreeMap[activeObject];
  if (tree == NULL || tree->isEmpty ()) {
	file.close ();
	statusBar->showMessage ("No data to export !", 0);
	return false;
  }
  tree->writeHFString (stream);
  file.close ();
  statusBar->showMessage ("HyperFun Export to " + fileName + " successful."  , 0);
  return true;
}


/////////////// ACCESS FUNCTIONS //////////////////////////////////////////

ToolInterface * createTool () {
  return new CoTeXTool;
}

TOOL_API void destroyTool (ToolInterface * tool) {
  delete tool;
}

const char * getToolName () {
  return "CoTeX";
}

// Some Emacs-Hints -- please don't remove:
//
//  Local Variables:
//  mode:C++
//  tab-width:4
//  End:
