#include <assert.h>
#include <math.h>
#include <vector>

#include "Rbf.h"
#include "LinearSolver.h"

static const int RBF_CONSTRAINT = 8; ///< 8 BBox exterior constraints
static const float RBF_DEFAULT_BBOX_SCALE = 1.2;
static const float RBF_DEFAULT_BBOX_VALUE = 30.0;

using namespace std;

Rbf::Rbf (QString n, RbfSet * p, TextureInterface * tex)
  : PrimitiveInterface (n, tex), rbfPointSet (p), results (NULL), bBoxMinMax (0, 0, 0, 0, 0, 0){
  assert (tex != NULL && p->size () >= RBF_MIN_POINTSET_SIZE 
	  && p->size () < RBF_MAX_POINTSET_SIZE && p != NULL);
  
  init ();
}

Rbf::Rbf (Rbf &rbf) : PrimitiveInterface (rbf.getName (), rbf.getTexture ()), 
		      bBoxMinMax (0, 0, 0, 0, 0, 0) {
  rbfPointSet = new RbfSet (*rbf.getRbfPointSet ());
  init ();
}

Rbf& Rbf::operator= (Rbf & rbf){
  delete rbfPointSet;
  rbfPointSet = rbf.getRbfPointSet ();
  init ();
  return (*this);
}

// ***************
// public methods
// ***************

void Rbf::setRbfPointSet (RbfSet * newRbfPointSet) {
  assert (newRbfPointSet != NULL && newRbfPointSet->size () >= 3);
  
  delete rbfPointSet;
  rbfPointSet = newRbfPointSet;
  init ();
}


double Rbf::eval (const Vector3D& position) {
  double result = 0.0;
  double value;

  unsigned int size = rbfPointSet->size ();
  for (unsigned int i = 0; i < size ; i++)
    {
      Vector3D tmp = (*rbfPointSet)[i];
      value = phi (tmp[0] - position[0], tmp[1] - position[1],
		  tmp[2] - position[2]);
      result += results[i] * value;
    }
  return (result + results[size] + position[0] * results[size+1] +
    position[1] * results[size+2] + position[2] * results[size+3]);
}


// **********************
// Interface HyperFunable
// **********************

void Rbf::writeHFString (QTextStream & outStream) {
  outStream << "my_model(x[3],a[1],s[3])" << endl << "{" << endl;
  writeDefArray (outStream);
  writeDefBody (outStream);
  outStream << endl << "}" << endl;
}


void Rbf::writeDefArray (QTextStream & outStream) {
  outStream << "array vect_" << getName () << "[" 
	    << rbfPointSet->size () << "];" << endl;
  outStream << "array results_" << getName () 
	    << "[" << (rbfPointSet->size () + 4) <<"];" << endl << endl;
}


void Rbf::writeDefBody (QTextStream & outStream) {
  QString vect_name = QString ("vect_") + getName ();
  QString results_name = QString ("results_") + getName ();
  QString color_name; 
  unsigned int i, j, pointSize, resultsSize;;
  
  if (getFather () != NULL) 
    color_name = "color_" + getName ();
  else 
    color_name = "s";
  
  // *****************
  // Writes the points
  // *****************
  
  outStream << vect_name << " = [";
  pointSize = rbfPointSet->size () - 1;
  for (i = 0 ; i < pointSize ; i++) {
    for (j = 0; j < 3; j++)
      outStream << (*rbfPointSet)[i][j] << ", ";
    outStream << endl;
  }
  outStream << (*rbfPointSet)[i][0] << ", " << (*rbfPointSet)[i][1] 
            << ", " << (*rbfPointSet)[i][2] << "];" << endl << endl; 
  
  
  // ******************
  // Writes the results
  // ******************
  
  outStream << results_name << " = [";
  resultsSize = rbfPointSet->size () + 3;
  for (i = 0 ; i < resultsSize ; i++) {
    outStream << results[i] << ", ";
    if (i % 5 == 0) // for presentation purpose
      outStream << endl;
  }
  outStream << results[i] << "];" << endl << endl;

  // ******************
  // Writes the texture
  // ******************
  
  outStream << color_name << " = ";
  getTexture ()->writeDefBody (outStream);
  outStream << ";" << endl << endl;  
  
  
  outStream << getName () << " = hfRbf (x, " << vect_name << ", " << results_name << ");" << endl << endl;
}

// *****************
// Interface Savable
// *****************

void Rbf::save (QTextStream &stream) {
  int size = rbfPointSet->size ();
  size -= RBF_CONSTRAINT;
  stream << "primitive rbf " <<  getName () << " " << size << " ";
  for(int i = 0 ; i < size ; i++)
    for (int j = 0 ; j < 3 ; j++)
      stream << (*rbfPointSet)[i][j] << " ";
  stream << " " << getTexture ()->getName ();
} 

void Rbf::load (QTextStream &stream) {
  QString n;
  int size;
  float x, y, z;

  delete rbfPointSet;
  rbfPointSet = new RbfSet ();
  stream >> n >> size;
  rbfPointSet->resize (size);
  for(int i = 0 ; i < size ; i++){
    stream >> x >> y >> z ;
    (*rbfPointSet)[i] = Vector3D (x, y, z);
  }
  setName (n);
  init ();
}


// **************
// Private method
// **************


void Rbf::init () {
  rbfReconstruction();
  if (results != NULL)
    delete [] results;
  results = new double[rbfPointSet->size()+4];
  solve();
}

BBoxMinMax Rbf::computeBBoxMinMax (){
  double xmin, xmax, ymin, ymax, zmin, zmax;
  unsigned int k;
  
  // **********************
  // Computing bounding box
  // **********************
  xmin = xmax = (*rbfPointSet)[0][0];
  ymin = ymax = (*rbfPointSet)[0][1];
  zmin = zmax = (*rbfPointSet)[0][2];

  for (k = 1; k < rbfPointSet->size (); k++) {
    if ((*rbfPointSet)[k][0] < xmin)
      xmin = (*rbfPointSet)[k][0];
    if ((*rbfPointSet)[k][1] < ymin)
      ymin = (*rbfPointSet)[k][1];
    if ((*rbfPointSet)[k][2] < zmin)
      zmin = (*rbfPointSet)[k][2];
    if ((*rbfPointSet)[k][0] > xmax)
      xmax = (*rbfPointSet)[k][0];
    if ((*rbfPointSet)[k][1] > ymax)
      ymax = (*rbfPointSet)[k][1];
    if ((*rbfPointSet)[k][2] > zmax)
      zmax = (*rbfPointSet)[k][2];
  }
  bBoxMinMax = BBoxMinMax (xmin - RBF_DEFAULT_BBOX_VALUE,
			   ymin - RBF_DEFAULT_BBOX_VALUE,
			   zmin - RBF_DEFAULT_BBOX_VALUE,
			   xmax + RBF_DEFAULT_BBOX_VALUE,
			   ymax + RBF_DEFAULT_BBOX_VALUE,
			   zmax + RBF_DEFAULT_BBOX_VALUE);
  bBoxMinMax.scale (RBF_DEFAULT_BBOX_SCALE);
  return bBoxMinMax;
}


double Rbf::phi (double x, double y, double z) {
  double r = sqrt (x * x + y * y + z * z); 
  return  r * r * r;
}

void Rbf::rbfReconstruction () {
  unsigned int size;
  vector<double> minMax;

  size = rbfPointSet->size ();
  
  computeBBoxMinMax ();
  minMax = bBoxMinMax.getMinMax ();
   
  rbfPointSet->resize (size + RBF_CONSTRAINT);
  (*rbfPointSet)[size] = Vector3D (minMax[3], minMax[4], minMax[5]);
  (*rbfPointSet)[size + 1] = Vector3D (minMax[3], minMax[4], minMax[2]);	
  (*rbfPointSet)[size + 2] = Vector3D (minMax[3], minMax[1], minMax[5]);	
  (*rbfPointSet)[size + 3] = Vector3D (minMax[3], minMax[1], minMax[2]);
  (*rbfPointSet)[size + 4] = Vector3D (minMax[0], minMax[4], minMax[5]);	
  (*rbfPointSet)[size + 5] = Vector3D (minMax[0], minMax[4], minMax[2]);	
  (*rbfPointSet)[size + 6] = Vector3D (minMax[0], minMax[1], minMax[5]);	
  (*rbfPointSet)[size + 7] = Vector3D (minMax[0], minMax[1], minMax[2]);	
}

void Rbf::solve () {
  int size = rbfPointSet->size ();
  double *a = new double[(size + 4) * (size + 4)];
  double *r = new double[size + 4];
  assert (a != 0);
  
  for (int j = 1; j <= size + 4; j++) {
      for (int i = 1; i <= size + 4; i++)	{
	double &aCell = a[(j - 1) * (size + 4) + (i - 1)];
	if (i <= size && j <= size) {	
	  Vector3D tmp1 = (*rbfPointSet)[j - 1];
	  Vector3D tmp2 = (*rbfPointSet)[i - 1];
	  aCell = phi (tmp1[0] - tmp2[0], tmp1[1] - tmp2[1], tmp1[2] - tmp2[2]);
	  continue;
	}
	  
	else if (i > size && i <= size + 4 && j <= size && i != j) {
	  Vector3D tmp1 = (*rbfPointSet)[j - 1];
	  if (i == size + 1) aCell = 1;
	  if (i == size + 2) aCell = tmp1[0];
	  if (i == size + 3) aCell = tmp1[1];
	  if (i == size + 4) aCell = tmp1[2];
	  continue;
	}
	else if (i <= size && i == j) {
	  aCell = phi (0,0,0);
	  continue;
	}
	else if (i <= size && j > size && j <= size + 4) {
	  Vector3D tmp1 = (*rbfPointSet)[i - 1];
	  if (j == size + 1) aCell = 1;
	  if (j == size + 2) aCell = tmp1[0];
	  if (j == size + 3) aCell = tmp1[1];
	  if (j == size + 4) aCell = tmp1[2];
	  continue;
	}
	else {
	  aCell = 0;
	}
      }
  }
  for (int i = 1; i <= size + 4; i++) {
    if (i <= size - RBF_CONSTRAINT)
      r[i - 1] = RBF_SURFACE_ISOVALUE;  // could be the regularization parameter lambda
    else
      r[i - 1] = RBF_NORMAL_ISOVALUE;
    if (i > size) r[i - 1] = 0;
  }
  LinearSolver::solve (size + 4, size + 4, a, r, results);
  delete [] a;
  delete [] r;
}
