#include <qbuttongroup.h>
#include <qpushbutton.h>
#include <qlistbox.h>
#include <qlayout.h>
#include <qwidgetstack.h>
#include <qpixmap.h>
#include <qpainter.h>
#include <qinputdialog.h>
#include <assert.h>

#include "TextureConfigurationWidget.h"
#include "ColorConfigurationWidget.h"
#include "PerlinNoiseConfigurationWidget.h"


enum TextureIDs {
  SIMPLE_COLOR_ID,
  PERLIN_NOISE_ID
};

static const QSize TEXTURE_WIDGET_SIZE = QSize (340, 180);
static const QSize TEXTURE_LISTBOX_SIZE = QSize (160, 170);
static const QSize TEXTURE_WIDGETSTACK_SIZE = QSize (220, 150);
static const QSize TEXTURE_PUSHBUTTON_SIZE = QSize (80, 20);
static const unsigned int TEXTURE_SPACING = 10;

TextureConfigurationWidget::TextureConfigurationWidget (TextureInterface::TextureMap * texMap,
							QWidget * parent, 
							const char * name, 
							WFlags f) 
  : QWidget (parent, name, f), texturesMap (texMap) {

  resizeWidget (this, TEXTURE_WIDGET_SIZE);
  
  QFont font (this->font ());
  font.setPointSize (COTEX_FONT_SIZE);
  this->setFont (font); 
 
  texturesListBox = new QListBox (this);
  resizeWidget (texturesListBox, TEXTURE_LISTBOX_SIZE);
  insertInTexturesList (QPixmap ("Resources/CoTeXColorIcon.png"), "simple color", SIMPLE_COLOR_ID);
  insertInTexturesList (QPixmap ("Resources/CoTeXPerlinNoiseIcon.png"), "perlin noise", PERLIN_NOISE_ID); 
  texturesListBox->setSelected (SIMPLE_COLOR_ID, TRUE); // by default
  connect (texturesListBox, 
	   SIGNAL (highlighted (int)),
	   this, 
	   SLOT (aboutTexture (int)));

  
  colorConfigurationWidget = new ColorConfigurationWidget;
  perlinNoiseConfigurationWidget = new PerlinNoiseConfigurationWidget;
  configurationWidgetMap[SIMPLE_COLOR_ID] = colorConfigurationWidget;
  configurationWidgetMap[PERLIN_NOISE_ID] = perlinNoiseConfigurationWidget;

  texturesConfigurationWidgetStack = new QWidgetStack (this);
  resizeWidget (texturesConfigurationWidgetStack, TEXTURE_WIDGETSTACK_SIZE);
  AbstractConfigurationWidgetMap::Iterator it;
  for (it = configurationWidgetMap.begin (); 
       it != configurationWidgetMap.end ();
       it++)
    texturesConfigurationWidgetStack->addWidget (it.data (), it.key ());
  texturesConfigurationWidgetStack->raiseWidget (colorConfigurationWidget); // by default
  
  QPushButton * okPushButton = createPushButton ("ok", this, TEXTURE_PUSHBUTTON_SIZE);
  connect (okPushButton, SIGNAL (clicked()), this, SLOT (okSlot()));

  QPushButton * cancelPushButton = createPushButton ("cancel", this, TEXTURE_PUSHBUTTON_SIZE);
  connect (cancelPushButton, SIGNAL (clicked()), this, SLOT (hide()));

  QHBoxLayout * buttonsHBoxLayout = new QHBoxLayout ();
  buttonsHBoxLayout->addWidget (okPushButton);
  buttonsHBoxLayout->addWidget (cancelPushButton);

  QVBoxLayout * rightSideVBoxLayout = new QVBoxLayout ();
  rightSideVBoxLayout->setSpacing (TEXTURE_SPACING);
  rightSideVBoxLayout->addWidget (texturesConfigurationWidgetStack);
  rightSideVBoxLayout->addLayout (buttonsHBoxLayout);

  QHBoxLayout * texturesHBoxLayout = new QHBoxLayout (this);
  texturesHBoxLayout->setSpacing (TEXTURE_SPACING);
  texturesHBoxLayout->addWidget (texturesListBox);
  texturesHBoxLayout->addLayout (rightSideVBoxLayout);
  texturesHBoxLayout->activate ();
}

void TextureConfigurationWidget::insertTexture (SimpleColorTexture * tex) {
  assert (tex != NULL);
  
  colorConfigurationWidget->conformTo (tex);
  QPixmap pix = colorConfigurationWidget->createPixmap ();
  
  addTexture (tex, pix);
}

void TextureConfigurationWidget::insertTexture (PerlinNoiseColorTexture * tex) {
  assert (tex != NULL);
  
  perlinNoiseConfigurationWidget->conformTo (tex);
  QPixmap pix = perlinNoiseConfigurationWidget->createPixmap ();
  
  addTexture (tex, pix);
}

// ***************
// Private methods
// ***************

void TextureConfigurationWidget::insertInTexturesList (QPixmap pix, const QString & nameTexture, const int textureId) {
  texturesListBox->insertItem (pix, nameTexture, textureId);
}

void TextureConfigurationWidget::resizeWidget (QWidget * widget, QSize size) {
  assert (widget != NULL);

  widget->resize (size);
  widget->setMinimumSize (size);
  widget->setMaximumSize (size);
}

void  TextureConfigurationWidget::addTexture (TextureInterface * tex, QPixmap pix) {
  texturesMap->insert (tex->getName (), tex);
  emit addNewTextureName (tex->getName (), pix);
}

QPushButton *  TextureConfigurationWidget::createPushButton (QString name,
							     QWidget * parent,
							     QSize size) {
  assert (parent != NULL);

  QPushButton * button = new QPushButton (name, parent);
  resizeWidget(button, size);
  return button;
}

// *************
// Private slots
// *************

void TextureConfigurationWidget::aboutTexture (int index) {
    texturesConfigurationWidgetStack->raiseWidget (configurationWidgetMap[index]);
}  
  
void TextureConfigurationWidget::okSlot () {  
  QPixmap pix (12, 12);
  TextureInterface * newTexture; 
  int textureSelectedId = texturesListBox->currentItem ();
  
  newTexture = configurationWidgetMap[textureSelectedId]->createTexture ();
  pix = configurationWidgetMap[textureSelectedId]->createPixmap ();
    
  // Textures names must be unique for a session
  bool ok = true; // if !ok, the texure must be destroyed
  while (texturesMap->contains (newTexture->getName ())
	 || newTexture->getName ().contains (' ') > 0
	 || newTexture->getName ().contains ('\t') > 0) {
    QString reallyNewFileName = QInputDialog::getText ("Enter a new name",
						       "Texture with this name already exists or this name contains space or tabulation. Enter a new name : ",
						       "newname", &ok, this);
    newTexture->setName (reallyNewFileName);
  }
  this->hide ();
  if (!ok 
      || (newTexture->getName ()).isNull ()
      || (newTexture->getName ()).isEmpty ())
    delete newTexture;
  else {
    texturesMap->insert (newTexture->getName (), newTexture);
    emit addNewTextureName (newTexture->getName (), pix);
  }
}




